import pkg_resources
import colorama

__all__ = ['BANNER']

P = colorama.Fore.MAGENTA + colorama.Style.BRIGHT
G = colorama.Fore.GREEN + colorama.Style.NORMAL
R = colorama.Style.RESET_ALL
W = colorama.Fore.WHITE + colorama.Style.NORMAL

distribution = pkg_resources.get_distribution('brainslug')
version = f'v{distribution.version}'

BANNER = (f"""
{P}  ______            __      {G} _______ __
{P} |   __ .----.---.-|__.-----{G}|     __|  .--.--.-----.
{P} |   __ |   _|  _  |  |     {G}|__     |  |  |  |  _  |
{P} |______|__| |___._|__|__|__{G}|_______|__|_____|___  |"""
          f"              {W}\n{version: >42}{G}   |_____|{R}").strip()



import asyncio
import functools

from concurrent.futures import ThreadPoolExecutor

from brainslug import web
from brainslug import runtime
from brainslug import ribosome


class Slug:
    def __init__(self, fn, spec):
        self.fn = fn
        self.spec = spec

    @classmethod
    def create(cls, **spec):
        return functools.partial(Slug, spec=spec)

    def __call__(self, *args, **kwargs):
        return self.fn(*args, **kwargs)

    async def run_in_thread(self, loop, fn):
        executor = ThreadPoolExecutor(max_workers=1)
        return await loop.run_in_executor(executor, fn)

    async def attach_resources(self, loop):
        resources = await runtime.wait_for_resources(loop,
                                                     runtime.AGENT_INFO,
                                                     self.spec)
        return functools.partial(self.fn, **resources)

    async def run(self):
        loop = asyncio.get_event_loop()
        prepared = await self.attach_resources(loop)
        return await self.run_in_thread(loop, prepared)


async def run_slug(slug):
    ribosome.load()
    runner = await web.run_web_server()
    try:
        return await slug.run()
    finally:
        await runner.cleanup()



import asyncio

from tinydb import TinyDB
from tinydb.storages import MemoryStorage


class AsyncTinyDB:
    """
    A tiny wrapper over TinyDB to allow notifications on insert.

    """
    def __init__(self):
        self._db = TinyDB(storage=MemoryStorage)
        self._new_channel = asyncio.Condition()

    async def wait_for_insert(self):
        async with self._new_channel:
            await self._new_channel.wait()

    async def insert(self, *args, **kwargs):
        ret = self._db.insert(*args, **kwargs)
        async with self._new_channel:
            self._new_channel.notify_all()
        return ret

    def search(self, *args, **kwargs):
        return self._db.search(*args, **kwargs)



import asyncio


class SyncedVar:
    """
    A variable synced using an asyncio.Event.

    A SyncedVar is a data-descriptor. Therefore is used as a class
    property.
        >>> class Foo:
        ...     bar = SyncedVar()

    On the instances...
        >>> foo = Foo()

    Setting the value is a synchronous operation:
        >>> foo.bar = 'myvalue'

    Getting the value is an asynchonous operation:
        >>> loop = asyncio.get_event_loop()
        >>> loop.run_until_complete(foo.bar)
        'myvalue'

    """
    def get_value(self, instance):
        """Return the stored value from the instance."""
        name = f"__syncvar_{id(self)}_value"
        if not hasattr(instance, name):
            setattr(instance, name, None)
        return getattr(instance, name)

    def set_value(self, instance, value):
        """Set the stored value from the instance."""
        return setattr(instance, f"__syncvar_{id(self)}_value", value)

    def get_event(self, instance):
        """Get the Event object used for synchronization."""
        name = f"__syncvar_{id(self)}_event"
        if not hasattr(instance, name):
            setattr(instance, name, asyncio.Event())
        return getattr(instance, name)

    def __get__(self, instance, owner):
        """
        Return a coroutine object who waits for the event before
        returning the value.

        .. important::
           The returned value has to be awaited.

        """
        async def wait_for_value():
            await self.get_event(instance).wait()
            self.get_event(instance).clear()
            return self.get_value(instance)
        return wait_for_value()

    def __set__(self, instance, value):
        """Set the value and the event."""
        self.set_value(instance, value)
        self.get_event(instance).set()


class Channel:
    """
    Manage the communication channel between an agent and one or more
    remote instances.

    """
    _code = SyncedVar()
    _result = SyncedVar()

    def __init__(self):
        self._eval_lock = asyncio.Lock()

    async def remote_eval(self, code):
        """
        Arrange the execution of `code` on the remote agent, eventually
        returning its result.

        """
        async with self._eval_lock:
            self._code = code
            return await self._result

    async def first_step(self):
        """Return the first code to be evaluated by the remote agent."""
        return await self._code

    async def next_step(self, last_result):
        """
        Receive the result of the last evaluation and return the next
        code to evaluate eventually.

        """
        self._result = last_result
        return await self._code



import asyncio
from uuid import uuid4
import os

from aiohttp import web
from tinydb import Query
import aiohttp_cors

from brainslug import runtime
from brainslug.ribosome import RIBOSOMES, Symbol
from brainslug.channel import Channel


def config_routes(app):
    app.add_routes([
        web.post('/channel/{__ribosome__}/{__key__}', channel_input),
        web.get('/launch/{__ribosome__}', get_launch),
        web.get('/boot/{__ribosome__}', get_boot),
        web.get('/boot/{__ribosome__}/{__key__}', get_boot)
    ])
    cors = aiohttp_cors.setup(app, defaults={
        "*": aiohttp_cors.ResourceOptions(
                allow_credentials=True,
                expose_headers="*",
                allow_headers="*",
            )
    })
    for resource in app.router.resources():
        cors.add(resource)


async def process_agent_request(ribosome, key, meta, last_result):
    Q = Query()
    try:
        [document] = runtime.AGENT_INFO.search(Q['__key__'] == key)
    except ValueError:
        document = {**meta,
                    '__key__': key,
                    '__ribosome__': ribosome,
                    '__channel__': Channel()}
        await runtime.AGENT_INFO.insert(document)
        return await document['__channel__'].first_step()
    else:
        return await document['__channel__'].next_step(last_result)


async def channel_input(request):
    ribosome = request.match_info['__ribosome__']
    key = request.match_info['__key__']
    meta = dict(request.rel_url.query)
    last_result = await request.read()

    next_code = await process_agent_request(ribosome, key, meta, last_result)

    return web.Response(body=next_code)


async def get_boot(request):
    ribosome = request.match_info['__ribosome__']
    key = request.match_info.get('__key__', str(uuid4()))
    try:
        boot = RIBOSOMES[Symbol((ribosome, 'boot'))]
    except KeyError as exc:
        raise web.HTTPNotFound() from exc
    else:
        url = request.url
        url = request.url.with_path(f'/channel/{ribosome}/{key}')
        url = url.with_query(request.url.query)
        return web.Response(body=boot(remote=None, url=url,
                                      **request.rel_url.query))


async def get_launch(request):
    ribosome = request.match_info['__ribosome__']
    try:
        launch = RIBOSOMES[Symbol((ribosome, 'launch'))]
    except KeyError as exc:
        raise web.HTTPNotFound() from exc
    else:
        url = request.url
        url = request.url.with_path(f'/boot/{ribosome}')
        url = url.with_query(request.url.query)
        return web.Response(body=launch(remote=None, url=url,
                                        **request.rel_url.query),
                            content_type="text/html")


async def run_web_server():
    app = web.Application(client_max_size=None)
    config_routes(app)

    runner = web.AppRunner(app)
    await runner.setup()

    server = web.TCPSite(runner,
                         os.environ.get('BRAINSLUG_HOST', '0.0.0.0'),
                         int(os.environ.get('BRAINSLUG_PORT', '8080')),
                         shutdown_timeout=0)
    asyncio.create_task(server.start())

    return runner



from pkg_resources import iter_entry_points

RIBOSOMES = dict()  #: Global ribosome registry


class Symbol(tuple):
    def __getattr__(self, name):
        if name.isidentifier():
            return Symbol(self + (name, ))
        else:
            raise AttributeError('Invalid attribute %r' % name)


def define(symbol):
    def _register(fn):
        def is_dunder(s):
            return s.startswith("__") and s.endswith("__")

        if any(is_dunder(part) for part in symbol):
            raise ValueError("Cannot define double underscore methods")
        elif symbol in RIBOSOMES:
            raise ValueError("Cannot define ribosome twice")
        else:
            return RIBOSOMES.setdefault(symbol, fn)

    return _register


def root(name):
    return Symbol((name, ))


def load():
    for entrypoint in iter_entry_points('brainslug.ribosomes'):
        entrypoint.load()



from argparse import Namespace
import asyncio
import sys

from tinydb import Query

from brainslug import _slug
from brainslug.banner import BANNER

__all__ = ['body', 'slug', 'run']

#: Used to query for resources
body = Query()

#: Decorator for slug definition
slug = _slug.Slug.create


def run(slug, local=False):
    print(BANNER)
    if local:
        resources = dict()
        for name in slug.spec:
            resources[name] = Namespace(**sys.modules)
        return slug.fn(**resources)
    else:
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(_slug.run_slug(slug))



from setuptools import setup, find_packages

VERSION = '0.0.2'

setup(name='brainslug',
      version=VERSION,
      description="Parasitic Computing Framework",
      long_description=open('README.rst', 'r').read(),
      classifiers=[
          "Development Status :: 3 - Alpha",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: Apache Software License",
          "Topic :: Internet",
          "Topic :: Security",
          "Topic :: Software Development :: Libraries :: Application Frameworks",
      ],
      keywords='parasitic computing framework',
      packages=find_packages(exclude=["tests", "docs", "poc"]),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
        'aiohttp==3.5.4',
        'tinydb==3.12.2',
        'colorama==0.4.1',
        'aiohttp-cors==0.7.0'
      ],
      entry_points={
        'brainslug.ribosomes': [
            'powershell = brainslug.ribosomes.powershell:powershell'
        ]
      })



from brainslug.database import AsyncTinyDB
from brainslug.remote import Remote

#: Global application state
AGENT_INFO = AsyncTinyDB()


def get_resources(loop, store, spec):
    resources = dict()

    for name, query in spec.items():
        found = store.search(query)
        if found:
            # TODO: Implement Many() to return a list
            agent_info = found[0]
            resources[name] = Remote.from_agent_info(loop, agent_info)
        else:
            return None

    return resources


async def wait_for_resources(loop, store, spec):
    resources = get_resources(loop, store, spec)
    while resources is None:
        await store.wait_for_insert()
        resources = get_resources(loop, store, spec)
    return resources



import asyncio

from brainslug import ribosome


class Remote:
    def __init__(self, symbol, eval=None, root=None):
        self.__symbol__ = symbol
        self.__eval__ = eval
        self.__root__ = self if root is None else root

    @classmethod
    def from_agent_info(cls, loop, agent_info):
        root = ribosome.root(agent_info['__ribosome__'])
        channel = agent_info['__channel__']

        def _run_threadsafe(code):
            return asyncio.run_coroutine_threadsafe(
                channel.remote_eval(code), loop).result()

        return cls(root, _run_threadsafe)

    def __getattr__(self, name):
        if name.isidentifier():
            return Remote(self.__symbol__ + (name, ), root=self.__root__)
        else:
            raise AttributeError('Invalid attribute %r' % name)

    def __ribosome__(self, *parts):
        return ribosome.RIBOSOMES[self.__symbol__ + parts]

    def __call__(self, *args, **kwargs):
        return self.__ribosome__()(self.__root__, *args, **kwargs)

    def __getitem__(self, name):
        return self.__ribosome__('getitem')(self.__root__, name)

    def __setitem__(self, name, value):
        return self.__ribosome__('setitem')(self.__root__, name, value)

    def __delitem__(self, name):
        return self.__ribosome__('delitem')(self.__root__, name)
